@@ -42,6 +42,22 @@ try:
 except ImportError:
     import collections as abc  # type: ignore
 
+try:
+    from itertools import accumulate
+except ImportError:
+
+    def accumulate(iterable, func=lambda a, b: a + b):
+        it = iter(iterable)
+        try:
+            total = next(it)
+        except StopIteration:
+            return
+        yield total
+        for element in it:
+            total = func(total, element)
+            yield total
+
+
 if False:
     from typing import Type, Tuple  # noqa
 
@@ -298,11 +314,19 @@ except ImportError:
     typing_root_type = ()  # type: Tuple[type, ...]
     ForwardRef = None
 else:
-    if hasattr(typing, "_Final"):  # new in Python 3.7
+    try:
+        # These types are new in Python 3.7, but also (partially) backported to the
+        # typing backport on PyPI.  Use if possible; or fall back to older names.
         typing_root_type = (typing._Final, typing._GenericAlias)  # type: ignore
         ForwardRef = typing.ForwardRef  # type: ignore
-    else:
+    except AttributeError:
         typing_root_type = (typing.TypingMeta, typing.TypeVar)  # type: ignore
+        try:
+            typing_root_type += (typing._Union,)  # type: ignore
+        except AttributeError:
+            # Under Python 3.5.0, we'll just give up... if users want strategies
+            # inferred from Union-typed attrs attributes they can try a newer Python.
+            pass
         ForwardRef = typing._ForwardRef  # type: ignore
 
 
@@ -374,6 +398,9 @@ if PY2:
         "co_cellvars",
     ]
 else:
+    # This field order is accurate for 3.5 - 3.7, but not 3.8 when a new field
+    # was added for positional-only arguments.  However it also added a .replace()
+    # method that we use instead of field indices, so they're fine as-is.
     CODE_FIELD_ORDER = [
         "co_argcount",
         "co_kwonlyargcount",
@@ -403,6 +430,12 @@ def update_code_location(code, newfile, newlineno):
     code you're probably here because it's broken something and now
     you're angry at me. Sorry.
     """
+    if hasattr(code, "replace"):
+        # Python 3.8 added positional-only params (PEP 570), and thus changed
+        # the layout of code objects.  In beta1, the `.replace()` method was
+        # added to facilitate future-proof code.  See BPO-37032 for details.
+        return code.replace(co_filename=newfile, co_firstlineno=newlineno)
+
     unpacked = [getattr(code, name) for name in CODE_FIELD_ORDER]
     unpacked[CODE_FIELD_ORDER.index("co_filename")] = newfile
     unpacked[CODE_FIELD_ORDER.index("co_firstlineno")] = newlineno
@@ -521,27 +554,6 @@ def implements_iterator(it):
     return it
 
 
-if PY3:
-    FileNotFoundError = FileNotFoundError  # type: Type[IOError]
-else:
-    FileNotFoundError = IOError
-
-# We need to know what sort of exception gets thrown when you try to write over
-# an existing file where you're not allowed to. This is rather less consistent
-# between versions than might be hoped.
-if PY3:
-    FileExistsError = FileExistsError  # type: Type[IOError]
-
-elif WINDOWS:
-    FileExistsError = WindowsError
-
-else:
-    # This doesn't happen in this case: We're not on windows and don't support
-    # the x flag because it's Python 2, so there are no places where this can
-    # be thrown.
-    FileExistsError = None
-
-
 # Under Python 2, math.floor and math.ceil return floats, which cannot
 # represent large integers - eg `float(2**53) == float(2**53 + 1)`.
 # We therefore implement them entirely in (long) integer operations.
