@@ -28,7 +28,7 @@ from hypothesis.internal.conjecture.data import (
     StopTest,
     bits_to_bytes,
 )
-from hypothesis.internal.conjecture.junkdrawer import IntList, uniform
+from hypothesis.internal.conjecture.junkdrawer import IntList
 
 
 class PreviouslyUnseenBehaviour(HypothesisException):
@@ -46,10 +46,23 @@ def inconsistent_generation():
 EMPTY = frozenset()
 
 
+@attr.s(slots=True)
+class Killed(object):
+    """Represents a transition to part of the tree which has been marked as
+    "killed", meaning we want to treat it as not worth exploring, so it will
+    be treated as if it were completely explored for the purposes of
+    exhaustion."""
+
+    next_node = attr.ib()
+
+
 @attr.s(slots=True)
 class Branch(object):
+    """Represents a transition where multiple choices can be made as to what
+    to drawn."""
+
     bit_length = attr.ib()
-    children = attr.ib()
+    children = attr.ib(repr=False)
 
     @property
     def max_children(self):
@@ -58,6 +71,8 @@ class Branch(object):
 
 @attr.s(slots=True, frozen=True)
 class Conclusion(object):
+    """Represents a transition to a finished state."""
+
     status = attr.ib()
     interesting_origin = attr.ib()
 
@@ -174,7 +189,7 @@ class TreeNode(object):
             and len(self.forced) == len(self.values)
             and self.transition is not None
         ):
-            if isinstance(self.transition, Conclusion):
+            if isinstance(self.transition, (Conclusion, Killed)):
                 self.is_exhausted = True
             elif len(self.transition.children) == self.transition.max_children:
                 self.is_exhausted = all(
@@ -196,31 +211,19 @@ class DataTree(object):
         described must have been fully explored."""
         return self.root.is_exhausted
 
-    def find_necessary_prefix_for_novelty(self):
-        """Finds a prefix that any novel example must start with.
-        This is currently only used for generate_novel_prefix, where
-        it allows us to significantly speed it up in the case where
-        we start with a very shallow tree.
+    def generate_novel_prefix(self, random):
+        """Generate a short random string that (after rewriting) is not
+        a prefix of any buffer previously added to the tree.
 
-        For example, suppose we had a test function that looked like:
-
-        .. code-block:: python
-
-            def test_function(data):
-                while data.draw_bits(1):
-                    pass
-
-        This has a unique example of size ``n`` for any ``n``, but we
-        only draw that example with probability ``2 ** (-n)`` through
-        random sampling, so we will very rapidly exhaust the search
-        space. By first searching to find the necessary sequence
-        that any novel example must satisfy, we can find novel
-        examples with probability 1 instead.
+        The resulting prefix is essentially arbitrary - it would be nice
+        for it to be uniform at random, but previous attempts to do that
+        have proven too expensive.
         """
-        necessary_prefix = bytearray()
+        assert not self.is_exhausted
+        novel_prefix = bytearray()
 
         def append_int(n_bits, value):
-            necessary_prefix.extend(int_to_bytes(value, bits_to_bytes(n_bits)))
+            novel_prefix.extend(int_to_bytes(value, bits_to_bytes(n_bits)))
 
         current_node = self.root
         while True:
@@ -231,56 +234,43 @@ class DataTree(object):
                 if i in current_node.forced:
                     append_int(n_bits, value)
                 else:
+                    while True:
+                        k = random.getrandbits(n_bits)
+                        if k != value:
+                            append_int(n_bits, k)
+                            break
                     # We've now found a value that is allowed to
                     # vary, so what follows is not fixed.
-                    return hbytes(necessary_prefix)
+                    return hbytes(novel_prefix)
             else:
-                assert not isinstance(current_node.transition, Conclusion)
+                assert not isinstance(current_node.transition, (Conclusion, Killed))
                 if current_node.transition is None:
-                    return hbytes(necessary_prefix)
+                    return hbytes(novel_prefix)
                 branch = current_node.transition
                 assert isinstance(branch, Branch)
-                if len(branch.children) < branch.max_children:
-                    return hbytes(necessary_prefix)
-                else:
-                    choices = [
-                        (k, v) for k, v in branch.children.items() if not v.is_exhausted
-                    ]
-                    assert len(choices) > 0
-                    if len(choices) == 1:
-                        k, v = choices[0]
-                        append_int(branch.bit_length, k)
-                        current_node = v
-                    else:
-                        return hbytes(necessary_prefix)
+                n_bits = branch.bit_length
 
-    def generate_novel_prefix(self, random):
-        """Generate a short random string that (after rewriting) is not
-        a prefix of any buffer previously added to the tree.
-
-        This is logically equivalent to generating the test case uniformly
-        at random and returning the first point at which we hit unknown
-        territory, but with an optimisation for the only common case where
-        that would be inefficient.
-        """
-        assert not self.is_exhausted
-
-        initial = self.find_necessary_prefix_for_novelty()
-
-        while True:
-
-            def draw_bytes(data, n):
-                i = data.index
-                if i < len(initial):
-                    return initial[i : i + n]
-                else:
-                    return uniform(random, n)
-
-            data = ConjectureData(draw_bytes=draw_bytes, max_length=float("inf"))
-            try:
-                self.simulate_test_function(data)
-            except PreviouslyUnseenBehaviour:
-                return hbytes(data.buffer)
+                check_counter = 0
+                while True:
+                    k = random.getrandbits(n_bits)
+                    try:
+                        child = branch.children[k]
+                    except KeyError:
+                        append_int(n_bits, k)
+                        return hbytes(novel_prefix)
+                    if not child.is_exhausted:
+                        append_int(n_bits, k)
+                        current_node = child
+                        break
+                    check_counter += 1
+                    # We don't expect this assertion to ever fire, but coverage
+                    # wants the loop inside to run if you have branch checking
+                    # on, hence the pragma.
+                    assert (  # pragma: no cover
+                        check_counter != 1000
+                        or len(branch.children) < (2 ** n_bits)
+                        or any([not v.is_exhausted for v in branch.children.values()])
+                    )
 
     def rewrite(self, buffer):
         """Use previously seen ConjectureData objects to return a tuple of
@@ -317,12 +307,15 @@ class DataTree(object):
                     data.conclude_test(t.status, t.interesting_origin)
                 elif node.transition is None:
                     raise PreviouslyUnseenBehaviour()
-                else:
+                elif isinstance(node.transition, Branch):
                     v = data.draw_bits(node.transition.bit_length)
                     try:
                         node = node.transition.children[v]
                     except KeyError:
                         raise PreviouslyUnseenBehaviour()
+                else:
+                    assert isinstance(node.transition, Killed)
+                    node = node.transition.next_node
         except StopTest:
             pass
 
@@ -335,6 +328,7 @@ class TreeRecordingObserver(DataObserver):
         self.__current_node = tree.root
         self.__index_in_current_node = 0
         self.__trail = [self.__current_node]
+        self.__killed = False
 
     def draw_bits(self, n_bits, forced, value):
         i = self.__index_in_current_node
@@ -384,6 +378,27 @@ class TreeRecordingObserver(DataObserver):
         if self.__trail[-1] is not self.__current_node:
             self.__trail.append(self.__current_node)
 
+    def kill_branch(self):
+        """Mark this part of the tree as not worth re-exploring."""
+        if self.__killed:
+            return
+
+        self.__killed = True
+
+        if self.__index_in_current_node < len(self.__current_node.values) or (
+            self.__current_node.transition is not None
+            and not isinstance(self.__current_node.transition, Killed)
+        ):
+            inconsistent_generation()
+
+        if self.__current_node.transition is None:
+            self.__current_node.transition = Killed(TreeNode())
+            self.__update_exhausted()
+
+        self.__current_node = self.__current_node.transition.next_node
+        self.__index_in_current_node = 0
+        self.__trail.append(self.__current_node)
+
     def conclude_test(self, status, interesting_origin):
         """Says that ``status`` occurred at node ``node``. This updates the
         node if necessary and checks for consistency."""
@@ -416,6 +431,10 @@ class TreeRecordingObserver(DataObserver):
         node.check_exhausted()
         assert len(node.values) > 0 or node.check_exhausted()
 
+        if not self.__killed:
+            self.__update_exhausted()
+
+    def __update_exhausted(self):
         for t in reversed(self.__trail):
             # Any node we've traversed might have now become exhausted.
             # We check from the right. As soon as we hit a node that
